#pragma once
#include <string>

#include "glm/glm.hpp"
#include "glad/glad.h"

#include "common.h"
#include "gl/shader.h"
#include "gl/gl_util.h"

namespace detail {
#pragma pack(push, 1)
    struct KV6Vertex {
        glm::vec3 vertex;
        glm::vec3 color;
        glm::vec3 normal; // face normal
        glm::vec3 kv6norm; // kv6 normal
    };
#pragma pack(pop)
}

struct KV6Mesh {
    explicit KV6Mesh(const std::string &name);

    void draw() const {
        this->mesh.draw();
    }

    ace::gl::experimental::mesh<detail::KV6Vertex> mesh{ "3f,3f,3f,3f" };

    int32_t xsiz, ysiz, zsiz, num_voxels;
    float xpiv, ypiv, zpiv;
};


struct KV6 {
    KV6(KV6Mesh *mesh, float scale=0.1f): scale(scale), rotation(0), position(0),
                                          lighting_rotation(0), mesh(mesh) {
    }

    bool sprhitscan(glm::vec3 p0, glm::vec3 v0, glm::vec3 *h);

    glm::mat4 get_model() const {
        return ace::model_matrix(position, rotation, scale);
    }

    glm::mat4 get_local_model() const {
        glm::vec3 rot(rotation);
        rot.y += 180;
        return ace::model_matrix(position, rot, scale);
    }

    void draw(ace::gl::ShaderProgram &s) const {
        const glm::mat4 model(this->get_model());
        s.uniform("model", model);
        s.uniform("normal_matrix", glm::mat3(transpose(inverse(model))));
        s.uniform("local", false);
        this->mesh->draw();
    }

    void draw_local(ace::gl::ShaderProgram &s) const {
        s.uniform("model", this->get_local_model());
        s.uniform("normal_matrix", glm::mat3(transpose(inverse(ace::model_matrix(position, lighting_rotation, scale))))); // for lighting and stuff. WHAT AM I DOING
        s.uniform("local", true);
        this->mesh->draw();
    }

    glm::vec3 scale, rotation, position, lighting_rotation;
    KV6Mesh *mesh;
};


struct KV6Manager {
    KV6Mesh *get(const std::string &name) {
        try {
            return &models.at(name);
        } catch (std::out_of_range &) {
//            auto x = std::make_unique<KV6Mesh>("kv6/" + name);
            return &models.emplace(name, ace::get_resource_path("kv6/" + name)).first->second;
        }
    }
private:
    std::unordered_map<std::string, KV6Mesh> models;
};


